import os,sys
import platform

from dbeif.utils.etracer import Tracer


from dbeif.utils.eif_exception   import FatalError, ConfigurationError, OperationalError
from dbeif.utils.claslo          import ClassLoader
from dbeif.csv_connection        import CSVConnection
from dbeif.utils.metrixeur       import recordement, metrixeur

ARCHITECTURE_PKG = 'dbeif'
__version__ = "1.1.0"

VALID_DATABASE_TYPES = ['Postgres', 'MSSQL']   # Future: add 'Oracle', 'SQLServer', 'MySQL' as available

class DBConnector:
    def __init__(self, config, logger):
        self.config = config
        self.log = logger

        conn_sections = []
        try:
            conn_sections = [section for section in config.topLevels() if section != 'Service']
            if len(conn_sections) == 0:
                problem = 'Config does not contain any connection sections'
                raise ConfigurationError(problem)
        except Exception:
            raise ConfigurationError('Config file lacks sufficient information for DBConnector')

        self.agicen_conn = None
        self.db_conn     = None
        self.db_name     = None
        self.db_type     = None

        db_entries = [name for name in conn_sections if name == 'DB']
        if db_entries:
            db_conf = self.config.topLevel(db_entries[0])
            self.db_name = db_conf.get('Name', None)
            self.db_type = db_conf.get('Type', None)
            if self.db_type:
                if self.db_type not in VALID_DATABASE_TYPES:
                    vdb_types = {vdb_type.lower() : vdb_type for vdb_type in VALID_DATABASE_TYPES}
                    if self.db_type.lower() in vdb_types.keys():
                        self.db_type = vdb_types[self.db_type.lower()]
                    elif  self.db_type.lower() == 'PostgreSQL'.lower():
                        self.db_type = 'Postgres'
                    else:
                        raise ConfigurationError('Unsupported DB Type: %s' % self.db_type)

        self.log.info("Agile Central DBConnector version %s" % (__version__))
        uname_fields = platform.uname()
        self.log.info("Python platform: %s %s" % (uname_fields[0], uname_fields[2]))
        self.log.info("Python  version: %s" % sys.version.split()[0])

        self.mr = None  # this will end up being a MetricsRecorder instance ...

        self.internalizeConfig(config)

        droid = ClassLoader()
        agicen_conn_class_name = config.connectionClassName('AgileCentral')
        if self.db_name and self.db_type:
            db_conn_class_name = config.connectionClassName('DB', prefix=self.db_type)
        else:
            db_conn_class_name = None

        try:
            self.agicen_conn_class = droid.loadConnectionClass(agicen_conn_class_name, pkgdir=ARCHITECTURE_PKG)
        except Exception as msg:
            raise FatalError('Unable to load AgileCentralConnection class, %s' % msg)
        try:
            if db_conn_class_name:
                self.db_conn_class = droid.loadConnectionClass(db_conn_class_name, pkgdir=ARCHITECTURE_PKG)
        except Exception as msg:
            raise FatalError('Unable to load %sConnection class, %s' % (self.db_type, msg))

        self.establishConnections()

        if not self.validate():  # basically just calls validate on both connection instances
            raise ConfigurationError("Validation failed")
        self.log.info("Initialization complete: Delegate connections operational, ready for scan/reflect ops")

    def internalizeConfig(self, config):
        self.agicen_conf = config.topLevel('AgileCentral')
        if self.db_name and self.db_type:
            self.db_conf = config.topLevel('DB')
        self.svc_conf = config.topLevel('Service')
        valid_config_items = ['Preview', 'LogLevel', 'SecurityLevel']
        svc_conf = config.topLevel('Service')
        invalid_config_items = [item for item in svc_conf.keys() if item not in valid_config_items]
        if invalid_config_items:
            problem = "Service section of the config contained these invalid entries: %s" % ", ".join(
                invalid_config_items)
            raise ConfigurationError(problem)


    def establishConnections(self):
        self.agicen_conn = self.agicen_conn_class(self.agicen_conf, self.log, self.mr)
        db_backend_version = ''
        if self.db_name and self.db_type:
            self.db_conn = self.db_conn_class(self.db_conf, self.log, self.mr)
            self.db_conn.connect()  # we do this before agicen_conn to be able to get the db backend version
            db_backend_version = self.db_conn.getBackendVersion()
            self.log.info("Connected to %s on %s:%s  database name: %s" % (db_backend_version, self.db_conn.server, self.db_conn.port, self.db_name))

        agicen_headers = {'name': 'Agile Central DBConnector',
                          'version': __version__,
                          'vendor': 'CA Technologies',
                          'other_version': 'No Database, CSV only - 1.1'
                        }
        if self.db_conn:
            agicen_headers['other_version'] = db_backend_version
            other_name    = self.db_conn_class.__name__
            other_version = self.db_conn.backend_version
        else:
            other_name    = 'CSV'
            other_version = '1.0'
        if getattr(self.agicen_conn, 'set_integration_header'):
            self.agicen_conn.set_integration_header(agicen_headers)
        self.agicen_conn.setSourceIdentification(other_name,other_version)
        self.agicen_conn.connect()  # so we can use it in our X-Rally-Integrations header items here

        delimiter = '\t' if self.db_type == 'MSSQL' else ','
        self.csv_conn = CSVConnection(self.log, delimiter)
        self.getFieldRemappings()


    def getFieldRemappings(self):
        self.remapping = {}
        if self.db_type == 'MSSQL':
            self.remapping = {'boolean' : {True : 1, False : 0}}


    def validate(self):
        """
            This calls the validate method on both the Agile Central and the DB connections
        """
        self.log.info("Connector validation starting")

        if not self.agicen_conn.validate():
            self.log.info("AgileCentralConnection validation failed")
            return False
        # if we support multiple projects we will revisit this:
        #if not self.agicen_conn.validateProjects(self.target_projects):
            #self.log.info("AgileCentralConnection validation for Projects failed")
            #return False
        self.log.info("AgileCentralConnection validation succeeded")
        #snarf the valid list of entities to be processed from self.agicen_conn
        entities = self.agicen_conn.entities
        if self.db_name:
            if not self.db_conn.validate(entities):   # pass the list of entities (aka table names) to the DBConnection for validation
                self.log.info("%sDBConnection validation failed" % self.db_type)
                return False
            else:
                self.log.info("DBConnection validation succeeded for %s" % self.db_name)

        self.log.info("Connector validation completed")

        return True

    def run(self):
        if 'DB' in self.config.topLevels():
            db_config = self.config.topLevel('DB')
        else:
            db_config = None

        populate_db = False
        if db_config and self.db_conn.populate and not self.config.topLevel('Service')['Preview']:
            populate_db = True
        trash_csv = True
        if db_config and self.db_conn.save_csv:
            trash_csv = False
        elif not db_config:
            trash_csv = False
        csv_dir_name = db_config.get('Name', 'default') if db_config else 'default'

        self.entity_scope = {'User' : {'Subscription' : False},
                             'Tags' : {'Workspace'    : False},
                            }

        stats = {}
        for entity in self.agicen_conn.entities:
            try:
                stats[entity] = self._processEntity(entity, csv_dir_name, populate_db, trash_csv)
            except Exception as e:
                exc_type, exc_msg, stack = sys.exc_info()
                self.log.error("%s in DBConnector.run processing entity: %s  -- %s" % (exc_type.__name__, entity, exc_msg))
                #self.log.write(stack)
                self.log.write(Tracer(e).getTrace())
                self.db_conn.disconnect()
                raise

        return True, stats

    def reportEntityQualifierCount(self, workspace, project, entity, scope_up, scope_down):
        qualifier_count = self.agicen_conn.getQualifierCount(entity, workspace, project, scope_up, scope_down)
        self.log.info('Preview mode: For the entity %s, there are %d qualifying items in %s : %s' % (entity, qualifier_count, workspace, project))


    @recordement(metrixeur, 'entity')
    def retrieveAndSaveAgileCentralData(self, entity):
        self.count = {}
        entity_schema     = None
        primary_key_field = ''
        project_scope_up = False
        for workspace in self.agicen_conn.workspaces:
            if entity in self.entity_scope and 'Subscription' in self.entity_scope[entity] and self.entity_scope[entity]['Subscription']:
                break
            workspace_name = workspace['Workspace']
            self.count[workspace_name] = 0
            projects = workspace.get('Projects', [None]) # konf returns sub elements
            for project in projects:
                if entity in self.entity_scope and 'Workspace' in self.entity_scope[entity] and self.entity_scope[entity]['Workspace']:
                    break
                project_scope_down = False

                if project and project.endswith('/*'):
                    project_scope_down = True
                    project = project.replace('/*', '')
                if self.config.topLevel('Service')['Preview']:
                    self.reportEntityQualifierCount(workspace_name, project, entity, project_scope_up, project_scope_down)
                    entity_schema = {}
                    continue

                entity_schema, primary_key_field, fields, entity_data, stats \
                    = self.agicen_conn.getAgileCentralData(workspace_name, project, entity, project_scope_up, project_scope_down, self.remapping)

                self.overall_stats[(workspace_name, project)] = stats

                if not entity_data:
                    self.log.warn('No qualifying results found for %s %s entity: %s' % (workspace, project, entity))
                    continue
                self.dumpDataToCSV(entity_data, fields)
                self.count[workspace_name] += stats['Number of Records']

                if entity in self.entity_scope and 'Workspace' in self.entity_scope[entity]:
                    self.entity_scope[entity]['Workspace'] = True
                    break

                if entity in self.entity_scope and 'Subscription' in self.entity_scope[entity]:
                    self.entity_scope[entity]['Subscription'] = True
                    break
        return entity_schema, primary_key_field

    def dumpDataToCSV(self, entity_data, fields):
        self.csv_conn.save(entity_data, fields)

    @recordement(metrixeur, 'entity')
    def copyToDatabase(self, entity, entity_schema, pk_field, csv_file):
        self.db_conn.prepareDatabase(entity)
        self.db_conn.createTableAndColumns(entity, entity_schema)
        self.db_conn.populateFromCSV(entity, pk_field, csv_file)

    def _processEntity(self, entity, csv_dir_name, populate_db, trash_csv):
        self.log.info("Processing type %s" % entity)
        csv_file = self.csv_conn.start(csv_dir_name, entity)
        self.log.debug("Preparing to write AC data to CSV file: %s" % csv_file )
        self.overall_stats = {}
        entity_schema, pk_field = self.retrieveAndSaveAgileCentralData(entity)

        if populate_db:
            try:
                self.copyToDatabase(entity, entity_schema, pk_field, csv_file)
            except Exception:
                raise
        if trash_csv:
            os.remove(csv_file)

        return self.overall_stats
